/*
 * Copyright 2016 Egbert Eich
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the authors
 * not be used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The authors make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "evdev.h"

#include <unistd.h>

#include <xf86.h>
#include <xf86Xinput.h>
#include <exevents.h>
#include <xorgVersion.h>

#include <X11/Xatom.h>
#include <evdev-properties.h>

#define DEBUG(x) x;

static Atom prop_ktouch        = 0; /* Kiosk touch emulation on/off property */
static Atom prop_ktouch_button = 0; /* Kiosk touch emulation button property */
static Atom prop_ktouch_button_delay = 0; /* Kiosk touch emulation button delay property */

#define KTOUCH_STATE_ACTIVE 1U << 0
#define KTOUCH_STATE_DELAY 1U << 1
#define KTOUCH_STATE_DELAY_MOTION 1U << 2

void
EvdevKioskTouchPreInit(InputInfoPtr pInfo)
{
    EvdevPtr pEvdev = (EvdevPtr)pInfo->private;
    int val;

    if (!(pEvdev->flags & EVDEV_ABSOLUTE_EVENTS) ||
	!(pEvdev->flags & EVDEV_BUTTON_EVENTS) ||
        (pEvdev->flags & EVDEV_RELATIVE_EVENTS) ||
        (pEvdev->flags & EVDEV_TOUCHPAD) ||
        !libevdev_has_event_code(pEvdev->dev, EV_ABS, ABS_X) ||
        !libevdev_has_event_code(pEvdev->dev, EV_ABS, ABS_Y)) {
        pEvdev->kioskTouch.mode = -1;
        return;
    }
    val = xf86SetIntOption(pInfo->options, "KioskTouchMode", 0);
    if (val < 0 || val > 2) {
        xf86Msg(X_WARNING, "%s: Invalid KioskTouchMode value: %d\n",
                pInfo->name, val);
        pEvdev->kioskTouch.mode = 0;
    } else {
        pEvdev->kioskTouch.mode = pEvdev->kioskTouch.mode_queued = val;
    }
    val = xf86SetIntOption(pInfo->options, "KioskTouchButton", 1);
    if (val > 255) {
         xf86Msg(X_WARNING, "%s: Invalid KioskTouchButton value: %d\n",
                pInfo->name, val);
         val = 0;
    }
    pEvdev->kioskTouch.button = pEvdev->kioskTouch.button_queued = val;
    val = xf86SetIntOption(pInfo->options, "KioskTouchButtonDelay", 0);
    pEvdev->kioskTouch.delay = pEvdev->kioskTouch.delay_queued = val;
    pEvdev->kioskTouch.state = 0;
    xf86Msg(X_INFO, "%s: KioskTouchpad mode initialized to %s - button: %d delay: %d\n",
            pInfo->name, (pEvdev->kioskTouch.mode == 0) ? "disabled" :
            (pEvdev->kioskTouch.mode == 1 ? "click-on-touch" : "click-on-release"),
            pEvdev->kioskTouch.button, pEvdev->kioskTouch.delay);
}

static void
EvdevKioskTouchSwitchQueued(EvdevPtr pEvdev)
{
    if (pEvdev->kioskTouch.mode != pEvdev->kioskTouch.mode_queued ||
        pEvdev->kioskTouch.button != pEvdev->kioskTouch.button_queued ||
        pEvdev->kioskTouch.delay != pEvdev->kioskTouch.delay_queued) {
        if (pEvdev->kioskTouch.state & KTOUCH_STATE_ACTIVE)
            return;
        if (pEvdev->mt_mask) {
            int i;
            for (i = 0; i < num_slots(pEvdev); i++) {
                if (pEvdev->slots[i].dirty)
                    return;
            }
        }
        pEvdev->kioskTouch.mode = pEvdev->kioskTouch.mode_queued;
        pEvdev->kioskTouch.button = pEvdev->kioskTouch.button_queued;
        pEvdev->kioskTouch.delay = pEvdev->kioskTouch.delay_queued;
    }
}

BOOL
EvdevKioskTouchFilterTouch(EvdevPtr pEvdev)
{
    EvdevKioskTouchSwitchQueued(pEvdev);
    return (pEvdev->kioskTouch.mode > 0) ? TRUE : FALSE;
}

static BOOL
EvdevKioskTouchStopButtonTimer(InputInfoPtr pInfo, BOOL queue)
{
    EvdevPtr     pEvdev = (EvdevPtr)pInfo->private;

    if ((pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY) == 0)
        return FALSE;

    pEvdev->kioskTouch.state &= ~KTOUCH_STATE_DELAY;
    if (queue) {
        if (pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY_MOTION) {
            EvdevPostButtonEvent(pInfo, pEvdev->kioskTouch.button, BUTTON_RELEASE);
            valuator_mask_copy(pEvdev->abs_vals, pEvdev->kioskTouch.vals);
            valuator_mask_zero(pEvdev->kioskTouch.vals);
        } else
            EvdevQueueButtonEvent(pInfo, pEvdev->kioskTouch.button, 0);
    } else {
        EvdevPostButtonEvent(pInfo, pEvdev->kioskTouch.button, BUTTON_RELEASE);
        if (pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY_MOTION) {
            xf86PostMotionEventM(pInfo->dev, Absolute, pEvdev->kioskTouch.vals);
            valuator_mask_zero(pEvdev->kioskTouch.vals);
        }
    }
    pEvdev->kioskTouch.state &= ~KTOUCH_STATE_DELAY_MOTION;
    return TRUE;
}

BOOL
EvdevKioskTouchFilterButton(InputInfoPtr pInfo, unsigned int button, int value)
{
    EvdevPtr pEvdev = (EvdevPtr)pInfo->private;

    EvdevKioskTouchSwitchQueued(pEvdev);
    if (button && button == pEvdev->kioskTouch.button) {
        if (value == 1)
            pEvdev->kioskTouch.state = KTOUCH_STATE_ACTIVE;
        else
            pEvdev->kioskTouch.state &= ~KTOUCH_STATE_ACTIVE;
        switch (pEvdev->kioskTouch.mode) {
        case 0:
            DEBUG((ErrorF("%s: mode 0 button %d value %d\n", \
                          __func__, button, value)))
            return FALSE;
        case 1:
            if (value == 1) {
                DEBUG((ErrorF("%s: Sending ButtonDown\n",__func__)))
                EvdevQueueButtonEvent(pInfo, button, 1);
                if (pEvdev->kioskTouch.delay > 0) {
                    pEvdev->kioskTouch.state |= KTOUCH_STATE_DELAY;
                    pEvdev->kioskTouch.expires = GetTimeInMillis ()
                        + pEvdev->kioskTouch.delay;
                } else {
                    DEBUG((ErrorF("%s: Sending ButtonUp\n",__func__)));
                    EvdevQueueButtonEvent(pInfo, button, 0);
                }
            } else if (value == 0) {
                DEBUG((ErrorF("%s: EvdevKioskTouchFilterButton: Filter Button UP\n",__func__)));
                if (pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY)
                    EvdevKioskTouchStopButtonTimer(pInfo, TRUE);
                pEvdev->kioskTouch.state = 0;
            }
            return TRUE;
        case 2:
            DEBUG((ErrorF("%s: mode 2 button %d value %d\n", \
                          __func__, button, value)));
            if (value == 1) {
                EvdevKioskTouchStopButtonTimer(pInfo, TRUE);
                return TRUE;
            } else if (value == 0) {
                EvdevQueueButtonEvent(pInfo, button, 1);
                if (pEvdev->kioskTouch.delay > 0) {
                    pEvdev->kioskTouch.state |= KTOUCH_STATE_DELAY;
                    pEvdev->kioskTouch.expires = GetTimeInMillis ()
                        + pEvdev->kioskTouch.delay;
                    return TRUE;
                } else
                    DEBUG((ErrorF("%s: Sending ButtonDown\n",__func__)));
           }
            return FALSE;
        default:
            return FALSE;
        }
    }

    return FALSE;
}

void EvdevKioskTouchMotionEvents(InputInfoPtr pInfo, ValuatorMask *vals)
{
    EvdevPtr pEvdev = (EvdevPtr)pInfo->private;

    if (pEvdev->kioskTouch.mode == 1) {
        int val;
        if ((pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY_MOTION) == 0) {
            valuator_mask_fetch(vals, 0, &pEvdev->kioskTouch.old_v[0]);
            valuator_mask_fetch(vals, 1, &pEvdev->kioskTouch.old_v[1]);
        }
        if (pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY) {
            if (valuator_mask_fetch(vals, 0, &val)) {
                valuator_mask_set(pEvdev->kioskTouch.vals, 0, val);
                valuator_mask_set(vals, 0, pEvdev->kioskTouch.old_v[0]);
            }
            if (valuator_mask_fetch(vals, 1, &val)) {
                valuator_mask_set(pEvdev->kioskTouch.vals, 1, val);
                valuator_mask_set(vals, 1, pEvdev->kioskTouch.old_v[1]);
            }
            pEvdev->kioskTouch.state |= KTOUCH_STATE_DELAY_MOTION;
        } else {
            valuator_mask_fetch(vals, 0, &pEvdev->kioskTouch.old_v[0]);
            valuator_mask_fetch(vals, 1, &pEvdev->kioskTouch.old_v[1]);
        }
    }
}

static void
EvdevKioskTouchWakeupHandler(pointer data,
                             int i,
                             pointer LastSelectMask)
{
    InputInfoPtr pInfo = (InputInfoPtr)data;
    EvdevPtr     pEvdev = (EvdevPtr)pInfo->private;
    int ms;

    if (pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY) {
        ms = pEvdev->kioskTouch.expires - GetTimeInMillis();
        if (ms <= 0) {
            int        sigstate;

            sigstate = xf86BlockSIGIO ();
            pEvdev->kioskTouch.state &= ~KTOUCH_STATE_DELAY;
            DEBUG((ErrorF("Delayed ButtonUP\n")));
            EvdevPostButtonEvent(pInfo, pEvdev->kioskTouch.button, BUTTON_RELEASE);
            if (pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY_MOTION) {
                xf86PostMotionEventM(pInfo->dev, Absolute, pEvdev->kioskTouch.vals);
                valuator_mask_zero(pEvdev->kioskTouch.vals);
                pEvdev->kioskTouch.state &= ~KTOUCH_STATE_DELAY_MOTION;
            }
            xf86UnblockSIGIO (sigstate);
        }
    }
}

static void
EvdevKioskTouchBlockHandler(pointer data,
                            struct timeval **waitTime,
                            pointer LastSelectMask)
{
    InputInfoPtr pInfo = (InputInfoPtr)data;
    EvdevPtr     pEvdev = (EvdevPtr)pInfo->private;
    int ms;

    if (pEvdev->kioskTouch.state & KTOUCH_STATE_DELAY) {
        ms = pEvdev->kioskTouch.expires - GetTimeInMillis ();
        if (ms <= 0)
            ms = 0;
        AdjustWaitForDelay (waitTime, ms);
    }
}

void
EvdevKioskTouchOn(InputInfoPtr pInfo)
{
    EvdevPtr     pEvdev = (EvdevPtr)pInfo->private;

    if (pEvdev->kioskTouch.mode == -1)
        return;

    RegisterBlockAndWakeupHandlers ((void (*)(void *, void *)) EvdevKioskTouchBlockHandler,
                                    (void (*)(void *, int)) EvdevKioskTouchWakeupHandler,
                                    (pointer)pInfo);
}

void
EvdevKioskTouchFinalize(InputInfoPtr pInfo)
{
    EvdevPtr     pEvdev = (EvdevPtr)pInfo->private;

    if (pEvdev->kioskTouch.mode == -1)
        return;

    RemoveBlockAndWakeupHandlers ((void (*)(void *, void *)) EvdevKioskTouchBlockHandler,
                                  (void (*)(void *, int)) EvdevKioskTouchWakeupHandler,
                                  (pointer)pInfo);
}

static int
EvdevKioskTouchSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
                           BOOL checkonly)
{
    InputInfoPtr pInfo  = dev->public.devicePrivate;
    EvdevPtr     pEvdev = pInfo->private;

    if (atom == prop_ktouch)
    {
        if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
            return BadMatch;

        if (!checkonly)
            pEvdev->kioskTouch.mode_queued = *((unsigned char*)val->data);
    } else if (atom == prop_ktouch_button) {

        if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
            return BadMatch;

        if (!checkonly)
            pEvdev->kioskTouch.button_queued = *((unsigned char*)val->data);
    } else if (atom == prop_ktouch_button_delay) {

        if (val->format != 16 || val->size != 1 || val->type != XA_INTEGER)
            return BadMatch;

        if (!checkonly)
            pEvdev->kioskTouch.delay_queued = *((CARD16*)val->data);
    }

    return Success;
}

void
EvdevKioskTouchInitProperty(DeviceIntPtr dev)
{
    InputInfoPtr pInfo  = dev->public.devicePrivate;
    EvdevPtr     pEvdev = pInfo->private;
    int          rc;

    /* This doesn't really belong to property handling, however it is convenient to do
     * here and avoids another function to be called */
    if (pEvdev->mt_mask && !libevdev_has_event_code(pEvdev->dev, EV_KEY, BTN_TOUCH)) {
        xf86Msg(X_WARNING, "%s: MultiTouch device has no BTN_TOUCH event: "
                "no Kiosk Mode support\n", pInfo->name);
        pEvdev->kioskTouch.mode = -1;
    } else {
        pEvdev->kioskTouch.vals = valuator_mask_new(2);
    }

    if (pEvdev->kioskTouch.mode < 0)
        return;

    prop_ktouch = MakeAtom(EVDEV_PROP_KIOSK_TOUCH, strlen(EVDEV_PROP_KIOSK_TOUCH), TRUE);
    rc = XIChangeDeviceProperty(dev, prop_ktouch, XA_INTEGER, 8,
                                PropModeReplace, 1,
                                &pEvdev->kioskTouch.mode,
                                FALSE);
    if (rc != Success)
        return;
    XISetDevicePropertyDeletable(dev, prop_ktouch, FALSE);

    prop_ktouch_button = MakeAtom(EVDEV_PROP_KIOSK_BUTTON, strlen(EVDEV_PROP_KIOSK_BUTTON),
                                 TRUE);
    rc = XIChangeDeviceProperty(dev, prop_ktouch_button, XA_INTEGER, 8,
                                PropModeReplace, 1,
                                &pEvdev->kioskTouch.button,
                                FALSE);
    if (rc != Success)
        return;
    XISetDevicePropertyDeletable(dev, prop_ktouch_button, FALSE);

    prop_ktouch_button_delay =
        MakeAtom(EVDEV_PROP_KIOSK_BUTTON_DELAY, strlen(EVDEV_PROP_KIOSK_BUTTON_DELAY), TRUE);
    rc = XIChangeDeviceProperty(dev, prop_ktouch_button_delay, XA_INTEGER, 16,
                                PropModeReplace, 1,
                                &pEvdev->kioskTouch.delay,
                                FALSE);
    if (rc != Success)
        return;
    XISetDevicePropertyDeletable(dev, prop_ktouch_button_delay, FALSE);

    XIRegisterPropertyHandler(dev, EvdevKioskTouchSetProperty, NULL, NULL);
}

void
EvdevKioskTouchClose(InputInfoPtr pInfo)
{
    EvdevPtr     pEvdev = pInfo->private;

    valuator_mask_free(&pEvdev->kioskTouch.vals);
}
